import re
import sys
import os
import math
import csv
import argparse
import numpy as np
import pandas as pd
from time import time
from itertools import product
from multiprocessing import Pool, cpu_count

"""Functions used for preprocessing
"""

def fsbc_seq_count(filename, output_path='./', bases=["A","C","G","T"]):
    """Calculate frequency of sequence. The sequence, frequency,
    and length of sequence are stored in 'count.csv'

    parameters
    ----------
    filename : string (/path/to/*.fasta or .fastq)
    output_path : string (default: './')
    bases : list (default: ["A","C","G","T"])
    
    Returns
    ----------
    """
    global N
    fd = open(output_path + 'count.csv','w')
    old_stdout = sys.stdout
    sys.stdout = fd
    with open(filename,'r') as f:
        words = []
        for line in f:
            match = re.match(r'[%s]+'%(str("".join(bases))), line)
            if match:
                words.append(line.strip())  
        word_freq = {}
        N = len(words)
        for c in words:
            word_freq[c] = word_freq.get(c,0) + 1
        lista = sorted(word_freq.items(), key=lambda x: x[1], reverse=True)
        print('{0},{1},{2}'.format("seq","F","L"))
        for (word, freq) in lista:
            print('{0},{1},{2}'.format(word, (freq), len(word))) 
    fd.close()
    sys.stdout = sys.__stdout__

"""Functions used for searching strings
"""

def fsbc_base_ratio(out_path='./'):
    """Get the ratio of the nucleic acid bases.

    Parameters
    ----------
    output_path : string (default: './')
    """
    global ratio_dict
    ratio_dict = {}
    with open(out_path, "r") as f:
        line = f.readline()
        line = f.readline()
        while line:
            line = line.strip()
            sequence, number, _ = line.split(",")
            base_count = {}
            for i in range(len(sequence)):
                base_count[sequence[i]] = base_count.get(sequence[i],0) + 1
            for key in base_count:
                ratio_dict[key] = ratio_dict.get(key, 0) + base_count[key]*float(number)
            line = f.readline()
    total = sum(ratio_dict.values())
    for key in ratio_dict:
        ratio_dict[key] = ratio_dict[key]/total


def fsbc_get_initial_subseq(DNA_bases, l_min):
    """Get the set of initial subsequences.

    Parameters
    ----------
    DNA_bases : list of string
    l_min : int

    Returns
    ----------
    res : list of string (substrings)
    """
    res = [''.join(element) for element in product(DNA_bases, repeat=l_min)]
    return res


def fsbc_check_self_overlap(subseq):
    """Check overlapping regions of the subsequence.

    Parameters
    ----------
    subseq : string

    Returns
    ---------
    res : list of strings
    """
    if(len(subseq))==1: return(None)
    else:
        res = []
        s1 = [subseq[-(i+1):len(subseq)] for i in range(len(subseq)-1)]
        s2 = [subseq[0:i+1] for i in range(len(subseq)-1)]
        for i in range(len(s1)):
            if s1[i] == s2[i]:
                res.append(s1[i])
    return res


def fsbc_calc_prob_subseq(subseq, ratio_dict):
    """Calculate the probability of subsequence Q.

    Parameters
    ----------
    subseq : string

    Returns
    ----------
    ratio_dict : dictionaly
    """
    res = 1
    for i in range(len(subseq)):
        res = res*ratio_dict[subseq[i]]
    return res


def fsbc_calc_prob_seq(subseq, ratio_dict, L=29):
    """Probability of sequences including specific subsequence (P_s,L).

    Parameters
    ----------
    subseq : string
    ratio_dict : dictionaly
    L : int (length of random region, default=29)

    Returns
    ----------
    P : list of int 
    """
    op = fsbc_check_self_overlap(subseq)
    l = len(subseq) # = l
    Q = fsbc_calc_prob_subseq(subseq, ratio_dict) # = Q
    P = [0]*(l-1)
    P.append(Q)
    for i in range(l,L):
        P_i = P[i-1] + Q*(1-P[i-l])
        for overlap in op:
            ll = i - l + len(overlap) 
            pp = fsbc_calc_prob_subseq(overlap, ratio_dict)
            P_i = P_i - (Q/pp)*(P[ll]-P[ll-1])
        P.append(P_i)
    return P

# Calculation of the number of sequences including the subsequence
### Too slowly
def sub_in_seq(seqs, subseq):
    """Check whether the sequences include the string.

    parameters 
    ----------
    seqs : dataframe
    subseq : string
    
    returns
    ----------
    res : bool
    """
    res = subseq in seqs
    return res


seq_r_f = pd.DataFrame({})


def sub_in_seq2(x):
    """Calculate the number of sequences including the subsequence.

    parameters
    ----------
    subseq : vector

    Returns
    res : int 
    ----------
    """
    global seq_r_f
    res = seq_r_f[np.vectorize(sub_in_seq)(seq_r_f.seq, x)].F.sum()
    return res 

def fsbc_calc_P_s(PL, len_ser):
    """Calculate expectation of sequences including specific subsequence (P_s).

    Parameters
    ----------
    PL : list 
    len_ser : series
    
    Returns
    ----------
    Ps : float
    """
    Ps = 0
    #import pdb; pdb.set_trace()
    for i, row in len_ser.T.iteritems():
        Ps += row*PL[i-1]
    return Ps

#@profile
def fsbc_calc_z(ss_df, ratio_dict, L_max, len_ser, threads):
    """Calculate Z-score of subsequence.

    Parameters
    ----------
    ss_df : DataFrame
    seq_r_f : DataFrame
    ratio_dict : dictionaly
    L_max : int
    len_ser : Series
    threads : int

    Returns
    ----------
    ss_df : DataFrame
    """
    global N
    PL_df = ss_df.subseq.apply(lambda x: fsbc_calc_prob_seq(x, ratio_dict, L_max))
    #import pdb; pdb.set_trace()
    ss_df['P_s'] = PL_df.apply(lambda x: fsbc_calc_P_s(x, len_ser)/N)
    if threads == 0:
        p = Pool()
    else:
        p = Pool(threads)
    ss_df['F'] = pd.Series(p.map(sub_in_seq2, ss_df.subseq))
    p.close()
    den = ss_df['P_s']*(1-ss_df['P_s'])*N
    den = den.apply(lambda x: math.sqrt(x))
    ss_df['Z'] = (ss_df['F'] - N*ss_df['P_s'])/den   
    return ss_df


def fsbc_norm(dataframe):
    """Normalize Z-score for string.

    Parameters
    ----------
    dataframe : dataframe
    
    Returns
    -------
    df_norm : dataframe
    """
    df_norm = dataframe.Z.apply(lambda x: (x-dataframe.Z.mean())/dataframe.Z.std())
    return df_norm


def fsbc_search_string(
        seqs, Fs, lmin, lmax, L_max, ratio_dict, DNA_bases, len_ser, threads):
    """Search subsequences and estimate strings.

    Parameters
    ----------
    seq_r_f : DataFrame
    lmin, lmax, L_max : int, int, int
    ratio_dict : dictionary
    DNA_bases : list of strings
    len_ser : Series
    threads : int
    
    Returns
    -----------
    df_M : DataFrame
    """
    global N
    print('%2d-string searching ...'%lmin)
    s0 = pd.DataFrame(np.array(fsbc_get_initial_subseq(DNA_bases, lmin)),
       columns=['subseq'])
    s0 = fsbc_calc_z(s0, ratio_dict, L_max, len_ser, threads) # calculate z-score
    s0 = s0[s0['Z']>0].reset_index(drop=True) # delete data if z<=0 # reset index
    s0['ZZ'] = fsbc_norm(s0) # normalization Z-score
    df_M = pd.DataFrame({'subseq':[],'P_s':[],'F':[],'Z':[],'ZZ':[]}) # initialization df_M
    df_M = df_M.append(s0)
    pd_version = int(pd.__version__.split(".")[1])
    if pd_version<17:
        for i in range(lmax-lmin):
            print('%2d-string searching ...'%(lmin+i+1))
            df_m = pd.DataFrame({'subseq':[],'P_s':[],'F':[],'Z':[]}) # initialization df_m
            for base in DNA_bases:
                s1 = pd.DataFrame({'subseq': s0['subseq'].map(lambda x: x+base)}, 
                   columns=['subseq'])
                s1 = fsbc_calc_z(s1, ratio_dict, L_max, len_ser, threads) # calculate z-score
                df_m = df_m.append(s1[s0['Z']<s1['Z']], ignore_index=True)
            if df_m.empty:
                break
            df_m['ZZ'] = fsbc_norm(df_m) # normalization df_m['Z']
            s0 = df_m.sort('subseq').reset_index(drop=True) # sort dataframe of string by subseq and reset index
            df_M = df_M.append(df_m) # update df_M by df_m
        df_M = df_M.sort('ZZ', ascending=False).reset_index(drop=True)
    else:
        for i in range(lmax-lmin):
            print('%2d-string searching ...'%(lmin+i+1))
            df_m = pd.DataFrame({'subseq':[],'P_s':[],'F':[],'Z':[]}) # initialization df_m
            for base in DNA_bases:
                s1 = pd.DataFrame({'subseq': s0['subseq'].map(lambda x: x+base)}, 
                   columns=['subseq'])
                s1 = fsbc_calc_z(s1, ratio_dict, L_max, len_ser, threads) # calculate z-score
                df_m = df_m.append(s1[s0['Z']<s1['Z']], ignore_index=True)
            if df_m.empty:
                break
            df_m['ZZ'] = fsbc_norm(df_m) # normalization df_m['Z']
            s0 = df_m.sort_values(by='subseq').reset_index(drop=True) # sort dataframe of string by subseq and reset index
            df_M = df_M.append(df_m) # update df_M by df_m
        df_M = df_M.sort_values(by='ZZ', ascending=False).reset_index(drop=True)
    return df_M

"""Functions used for clustering sequences
"""

def fsbc_cluster(df_m, df_s, out_path, ranking=1, min_c=0):
    """Generate clusters based on subsequence information.

    Parameters
    ----------
    df_m : DataFrame
    df_s : DataFrame
    out_path : string
    ranking : int (default=1)
    min_c : int (default=0)
    """
    if len(df_m) > 0 and len(df_s) > 0:
        m = df_m.subseq[0]
        rows = np.vectorize(lambda x: m in x)(df_s.seq)
        num = rows.sum()
        if num>min_c:
            res = pd.concat([df_s[rows].reset_index(drop=True), 
                pd.DataFrame({'string':[m]*num, 'ranking':[ranking]*num})], 
                axis=1)
            res.to_csv(out_path+"clustered.csv", header=False, index=False, 
                       mode='a')
            ranking += 1
        fsbc_cluster(
            df_m[np.vectorize(lambda x: m not in x)(df_m.subseq)].reset_index(drop=True), 
            df_s[[not(x) for x in rows]].reset_index(drop=True), 
            out_path, ranking)


def fsbc_cluster_forloop(df_m, df_s, out_path, ranking=1, min_c=0):
    """Generate clusters based on subsequence information.

    Parameters
    ----------
    df_m : DataFrame
    df_s : DataFrame
    out_path : string
    ranking : int (default=1)
    min_c : int (default=0)
    """
    os.makedirs(out_path, exist_ok=True)
    output_file = open(out_path+'/clustered.csv', 'w')
    output_file.write('seq,F,L,subseq,ranking\n')
    output_file.close()
    while len(df_m) > 0 and len(df_s) > 0:
        m = df_m.subseq[0]
        rows = np.vectorize(lambda x: m in x)(df_s.seq)
        num = rows.sum()
        if num > min_c:
            res = pd.concat([df_s[rows].reset_index(drop=True), 
                pd.DataFrame({'string':[m]*num, 'ranking':[ranking]*num})], 
                axis=1)
            res.to_csv(out_path+'/clustered.csv', header=False, index=False, 
                       mode='a')
            ranking += 1
            df_s = df_s[np.logical_not(rows)].reset_index(drop=True)
        df_m = df_m[np.vectorize(lambda x: m not in x)(df_m.subseq)].reset_index(drop=True)


"""main
"""
parser = argparse.ArgumentParser(
    prog='python3 FSBC.py',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=("""
    _____ ______ ____  ______  
   / ____/  ____/ __ )/ ____/  
  / /_   \__  \/ __  / /      
 / __/  ____) / /_/ / /___   
/_/    /_____/_____/\____/ 
Fast String-Based Clustering method for HT-SELEX data                                                             
FSBC processes a FASTA file or a seq file, estimates string\n
and make clusters with those strings.\n
For example:\n
if you implement string-based clustering for fasta file deriverd HT-SELEX;\n
>>> python3 FSBC.py -i ./candidate.fa -o ./result
>>> python3 FSBC.py -i ./Test/H1_R5_SELEX.sense_seq_f9.fastq -o ./Test/ -npre"""))

parser.add_argument("-i","--input",help="Input-file to be analyzed. Format accepted is FASTA or seq")
parser.add_argument("-o","--output",help="Path to output analyzed data, or 'count.csv', or 'count.csv' and 'strings.csv'")
parser.add_argument("-rna","--rnasequence",action="store_true",help="Uses RNA nt [AUCG] as letter list. Default letter list is [ATCG]")
parser.add_argument("-min","--minlen",type=int,help="Minimum length of subsequences. Default is 5",default=5)
parser.add_argument("-max","--maxlen",type=int,help="Maximum length of subsequences. Default is 10",default=10)
parser.add_argument("-nest","--not_estimation",action="store_true",help="Truncate one base from the one side at a time")
parser.add_argument("-nclus","--not_clustering",action="store_true",help="Truncate each one base of both sides")
parser.add_argument("-npre","--not_preprocess",action="store_true",help="Truncate arbitrary one base of the aptamer at a time")
parser.add_argument("-th","--thread",type=int,help="No. of threads used to search strings. Default is the maximum number of threads of your computer",default=0)
args = parser.parse_args()

if __name__=="__main__":
    print(pd.__version__)
    in_path = args.input
    out_path = args.output

    os.makedirs(out_path, exist_ok=True)
    if args.rnasequence:
        letter_list = ["A","U","C","G"]
    else:
        letter_list = ["A","T","C","G"]
    #####################
    ### Preprocessing ###
    #####################
    if args.not_preprocess:
        print('Skip preprocessing.')
        print('Loading %s'%(in_path))
        # seq_r_f = pd.read_csv(out_path+"count.csv")
        seq_r_f = pd.read_csv(in_path)
        N  = seq_r_f.F.sum()
        print('%d total sequences.\n%d unique sequences.\n'%(N, len(seq_r_f)))
    else: # Calculate frequency (sequence & frequency & length of sequence)
        print('Loading %s'%(in_path.replace('./','"')+'"'))
        fsbc_seq_count(in_path, out_path, letter_list)
        seq_r_f = pd.read_csv(out_path+"count.csv")
        print('%d total sequences.\n%d unique sequences.\n'%(N, len(seq_r_f)))
    #############################
    ### Estimation of strings ###
    #############################
    t1 = time()
    if args.not_estimation:
        print('Skip calculation of base ratio.')
        # df_M = pd.read_csv(out_path+'strings.csv')
        df_M = pd.read_csv(in_path)
    else: 
        print('Calculating base ratio ...')
        if args.not_preprocess:
            fsbc_base_ratio(in_path)
        else:
            fsbc_base_ratio(out_path+'count.csv')
        print('Base ratio: ',ratio_dict,'\n')
        lmin = args.minlen
        lmax = args.maxlen
        Lmax = max(seq_r_f['L'])
        print('Estimation of strings ...')      
        len_ser = pd.DataFrame([seq_r_f[seq_r_f.L==x].F.sum() for x in seq_r_f.L.unique()],index=seq_r_f.L.unique())  
        df_M = fsbc_search_string(
             seq_r_f.seq, seq_r_f.F, lmin, lmax, Lmax, ratio_dict, 
             list(ratio_dict.keys()), len_ser, args.thread)
        print('\nFinished !\nExecution time: {}'.format(time()-t1))
        df_M.to_csv(out_path+'strings.csv', index=False)
    print('No. of strings: %d'%len(df_M))
    ##################
    ### Clustering ###
    ##################
    if args.not_clustering:
        print('Skip the clustering process.')
    else:
        print('\nGenerating Cluster ...')
        t1 = time()
        # fsbc_cluster_v2(df_M, out_path)
        if os.path.isfile(out_path+'clustered.csv'):
            os.remove(out_path+'clustered.csv')
        # fsbc_cluster(df_M, seq_r_f, out_path)
        fsbc_cluster_forloop(df_M, seq_r_f, out_path)
        print('Cluster generated !\nExecution time: {}'.format(time()-t1))
    print('Process is in success!')
